"""A set of scripts for running BRER simulations using gmxapi.

Online documentation is hosted at https://kassonlab.github.io/brer_md/.

See Also
--------
Hays, J. M., Cafiso, D. S., & Kasson, P. M.
Hybrid Refinement of Heterogeneous Conformational Ensembles using Spectroscopic Data.
*The Journal of Physical Chemistry Letters.*
DOI: `10.1021/acs.jpclett.9b01407 <https://pubs.acs.org/doi/10.1021/acs.jpclett.9b01407>`__
"""
__all__ = ['__version__']

from ._version import __version__
